/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include <time.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <termios.h>
#include <dirent.h> 
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>

/* 1. Type Definition                           */
/************************************************/
typedef signed char         int_08t;
typedef signed short        int_16t;
typedef signed int          int_32t;
typedef signed long long    int_64t;
typedef unsigned char      uint_08t;
typedef unsigned short     uint_16t;
typedef unsigned int       uint_32t;
typedef unsigned long long uint_64t;

typedef uint_32t INST_TYPE;
typedef uint_64t ADDR_TYPE;
typedef uint_64t DATA_TYPE;


/* 2. Constant Definition                       */
/************************************************/
#define NAME "SimAlpha Release"
#define VER  "Version 1.1.24 2003-08-14"

#define DEF_FILE_NAME    "aout.txt"
#define BLOCK_SIZE       0x00002000 /* 8KB page */
#define BLOCK_TABLE_SIZE 0x00080000 /* page size*/
#define BLOCK_MASK       0x00001fff /* page mask*/
#define BLOCK_MASK_BIT   13         /* mask bit */
#define DATA_T_SIZE      8    /* size of data_t */ 

/** for AMASK op  and IMPLVER op **/
#define IMPLEMENTATION_VERSION 2   
#define CPU_FEATURE_MASK 0x0000000000000107ull 

#define FD_MAPPING_MAX 128      /* read/write  */
#define OSF_SIGCONTEXT_SIZE 568 /* in byte     */
#define OSF_FSTAT_SIZE       80 /* in byte     */
#define OSF_RUSAGE_SIZE      72 /* in byte     */
#define OPTION_MAX          30  /* max options */

/** for systemcall implementation              **/
/** I implement one used for SPEC CPU2000.     **/
/************************************************/
#define ASYS_exit                1
#define ASYS_read                3
#define ASYS_write               4
#define ASYS_close               6
#define ASYS_unlink             10
#define ASYS_brk                17
#define	ASYS_lseek	        19
#define ASYS_getpid             20
#define ASYS_getuid             24
#define ASYS_open               45
#define ASYS_getgid             47
#define ASYS_sigprocmask        48
#define ASYS_ioctl              54
#define ASYS_getpagesize        64
#define ASYS_stat               67
#define ASYS_mmap               71
#define ASYS_munmap             73
#define ASYS_fstat              91
#define ASYS_fcntl              92
#define	ASYS_sigreturn         103
#define	ASYS_gettimeofday      116
#define	ASYS_getrusage         117
#define	ASYS_rename            128
#define	ASYS_getrlimit         144
#define	ASYS_setrlimit         145
#define	ASYS_sigaction         156
#define	ASYS_getdirentries     159
#define ASYS_UNIMPLEMENT       226
#define ASYS_osf_getsysinfo    256
#define ASYS_osf_setsysinfo    257

/** Table C-6: Opcode Summary                  **/
/** by Alpha Architecture Handbook V.4         **/
/************************************************/
/** 0x00  PALcode instruction(CALL_PAL)        **/
/** 0x08  load address instruction             **/
/** 0x09  load address high instruction        **/
/** 0x10  Integer arithmetic instruction       **/
/** 0x11  Integer logical instruction          **/
/** 0x12  Integer shift instruction            **/
/** 0x13  Integer multiply instruction         **/
/** 0x14  Integer to floating-point register   **/
/** 0x15  VAX floating-point instruction       **/
/** 0x16  IEEE floating-point instruction      **/
/** 0x17  Floating-point instruction           **/
/** 0x18  Miscellaneous instruction            **/
/** 0x1a  Jump instruction                     **/
/** 0x1c  Floating-point to integer register   **/
/************************************************/
#define OP_PAL  0x00
#define OP_LDA  0x08
#define OP_LDAH 0x09
#define OP_INTA 0x10
#define OP_INTL 0x11
#define OP_INTS 0x12
#define OP_INTM 0x13
#define OP_ITFP 0x14
#define OP_FLTV 0x15
#define OP_FLTI 0x16
#define OP_FLTL 0x17
#define OP_MISC 0x18
#define OP_JSR  0x1a
#define OP_FPTI 0x1c

#define BIT0    0x0000000000000001ull
#define BIT1    0x0000000000000002ull
#define BIT2    0x0000000000000004ull
#define BIT3    0x0000000000000008ull
#define BIT4    0x0000000000000010ull
#define BIT5    0x0000000000000020ull
#define BIT6    0x0000000000000040ull
#define BIT7    0x0000000000000080ull
#define BIT8    0x0000000000000100ull
#define BIT9    0x0000000000000200ull
#define BIT10   0x0000000000000400ull
#define BIT11   0x0000000000000800ull
#define BIT12   0x0000000000001000ull
#define BIT15   0x0000000000008000ull
#define BIT20   0x0000000000100000ull
#define BIT31   0x0000000080000000ull
#define BIT63   0x8000000000000000ull

#define EXTND8  0xffffffffffffff00ull
#define EXTND16 0xffffffffffff0000ull
#define EXTND21 0xffffffffffe00000ull
#define EXTND32 0xffffffff00000000ull
#define MASK08  0x00000000000000ffull
#define MASK16  0x000000000000ffffull
#define MASK21  0x00000000001fffffull
#define MASK32  0x00000000ffffffffull

#define ZAP0    0xffffffffffffff00ull
#define ZAP1    0xffffffffffff00ffull
#define ZAP2    0xffffffffff00ffffull
#define ZAP3    0xffffffff00ffffffull
#define ZAP4    0xffffff00ffffffffull
#define ZAP5    0xffff00ffffffffffull
#define ZAP6    0xff00ffffffffffffull
#define ZAP7    0x00ffffffffffffffull

/** for instruction.cc **/
#define MSK1 0x3e
#define MSK2 0x3c
#define MSK3 0x38
#define MSK4 0x30


/* 4. structure and class definition            */
/************************************************/
/*** Used in syscall.c SYS_fstat ***/
struct osf_statbuf{ 
  uint_32t a_st_dev;
  uint_32t a_st_ino;
  uint_32t a_st_mode;
  uint_16t a_st_nlink;
  uint_16t a_pad0;
  uint_32t a_st_uid;
  uint_32t a_st_gid;
  uint_32t a_st_rdev;
  uint_32t a_pad1;
  uint_64t a_st_size;
  uint_32t a_st_atime;
  uint_32t a_st_spare1;
  uint_32t a_st_mtime;
  uint_32t a_st_spare2;
  uint_32t a_st_ctime;
  uint_32t a_st_spare3;
  uint_32t a_st_blksize;
  uint_32t a_st_blocks;
  uint_32t a_st_gennum;
  uint_32t a_st_spare4;
};

/************************************************/
class Env{
 public:
  class system_config      *sc;
  class evaluation_result  *e;
  class debug              *deb;
  class system_manager     *sys;
  class memory_system      *mem;
  class architecture_state *as;
};

/************************************************/
class system_config{
 public:
  char program_name[512];
  uint_64t end_tc;
  int verbose_interval;
  int debug_flag;
  int syscall_flag;
  int simple_result;
  int fd_adjust;
  int fd_stdin_flag;
  int fd_stdin; /* file descripter for STDIN */
  system_config(char *, char**);
};

/************************************************/
class evaluation_result{
 public:
  uint_64t retired_inst; 
  int used_memory_block; 
  time_t time_begin;     /* start time stamp */
  struct timeval tp;     /* start time stamp */
  struct timezone tzp;   /* start time stamp */
  evaluation_result();
};

/************************************************/
class data_t{
  uint_64t value;
 public:
  int cmov;      
  uint_64t ld();     
  int   st(uint_64t);
  int init(uint_64t);
};

/************************************************/
class architecture_state{
  Env *ev;
 public:
  data_t pc;        /* program counter      */
  data_t r[32];     /* general purpose regs */
  data_t f[32];     /* floating point  regs */
  uint_64t rr(int); /* register read        */
  architecture_state(Env*);
};

/************************************************/
class main_memory{
  Env *ev;
  data_t *block_table[BLOCK_TABLE_SIZE];
  data_t *allocblock(data_t *);
 public:
  void ld_8byte(data_t *, data_t *);
  void st_8byte(data_t *, data_t *, DATA_TYPE);
  main_memory(Env*);
};

/************************************************/
class memory_system{
  Env *ev;
  class main_memory *mm;
 public:
  void ld_8byte(data_t *, data_t *);
  void st_8byte(data_t *, data_t *, DATA_TYPE);
  void ld_inst(data_t *, INST_TYPE *);
  void ld_nbyte(int, data_t *, data_t *);
  void st_nbyte(int, data_t *, data_t *);
  ~memory_system();
  memory_system(Env*);
};

/************************************************/
class system_manager{
  Env *ev;
  /* File Descriptor Mapping Table */
  DATA_TYPE fd_table[FD_MAPPING_MAX]; 
  /* Unique Data for rduniq & wruniq inst */
  DATA_TYPE uniq; 
 public:
  int running;    /* chip is running ? */
  void execute_pal(class instruction*);
  int translate_fd(DATA_TYPE);
  system_manager(Env*);
};

/************************************************/
class debug{
  Env *ev;
 public:
  debug(Env*);
  void debugmode();
};

/************************************************/
class instruction{
  Env *ev;
  INST_TYPE ir; /* 32bit instruction code  */
  int Op;       /* Opcode field            */
  int RA;       /* Ra field of the inst    */
  int RB;       /* Rb field of the inst    */
  int RC;       /* Rc field of the inst    */
  int ST;       /* store inst ?            */
  int LD;       /* load inst ?             */
  int CM;       /* CMOV inst ?             */
  int LA;       /* load address inst ?     */
  int BR;       /* branch inst ?           */
  int Ai;       /* Rav is immediate ?      */
  int Bi;       /* Rbv is immediate ?      */
  int Af;       /* Rav from floating-reg ? */
  int Bf;       /* Rbv from floating-reg ? */
  int WF;       /* Write to the f-reg ?    */
  int WB;       /* Writeback reg index     */
  data_t  Cpc;  /* PC of the instruction   */
  data_t  Npc;  /* PC of the next inst.    */
  data_t  Imm;  /* immediate               */
  data_t  Adr;  /* load & store address    */
  data_t  Rav;  /* Ra                      */
  data_t  Rbv;  /* Rb                      */
  data_t  Rcv;  /* Rc                      */
 public:
  int Fetch(data_t*);
  int Fetch(data_t*, INST_TYPE);
  int Slot();
  int Rename();
  int Issue();
  int RegisterRead();
  int Execute(data_t*);
  int Memory();
  int WriteBack();
  INST_TYPE get_ir();
  instruction(Env*);
};

/************************************************/
class simple_chip{
 public:
  Env         *ev;
  instruction *p;
  simple_chip(char*, char**);
  ~simple_chip();
  int step();
};

/************************************************/
struct alpha_stat{  /*** for syscall.cc ***/
  uint_64t  a_st_dev;         
  uint_32t  a_st_ino;              
   int_32t  a___pad1;
  uint_32t  a_st_mode;            
  uint_32t  a_st_nlink;          
  uint_32t  a_st_uid;              
  uint_32t  a_st_gid;              
  uint_64t  a_st_rdev;             
   int_64t  a_st_size;             
   int_64t  a_st_atime;           
   int_64t  a_st_mtime;           
   int_64t  a_st_ctime;
  uint_32t  a_st_blocks;        
   int_32t  a___pad2;
  uint_32t  a_st_blksize;     
  uint_32t  a_st_flags;
  uint_32t  a_st_gen;
   int_32t  a___pad3;
   int_64t  a___unused[4];
};


/* 5. function prototypes                       */
/************************************************/
extern void set_st_data(INST_TYPE, data_t*,
                        data_t*, data_t*,
                        DATA_TYPE*);
extern void set_ld_data(INST_TYPE, data_t*,
                        data_t*, data_t*);
extern void print_evaluation_result(Env*);
extern void usage();
extern uint_64t map_s(unsigned int);
extern INST_TYPE cmov_ir_create(INST_TYPE);
extern int ALU(INST_TYPE, data_t*, 
	       data_t*, data_t*);
extern int BRU(INST_TYPE, data_t*, data_t*, 
	       data_t*, data_t*);
extern void option_from_file(char**, char*);

extern int execute_cmovb(instruction*, 
			 architecture_state*);
extern void house_keeper(Env*);
extern int nbyte(INST_TYPE);
extern "C" int simple_cpu(int, uint_64t*);
extern uint_64t get_hex(char*);
/************************************************/
